﻿using System;
using System.IO;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using Microsoft.WindowsAPICodePack.Shell;
using FarNet;

namespace Win7Libs
{
    [ModuleTool(Name = Windows7Libraries.Name, Options = ModuleToolOptions.Panels | ModuleToolOptions.Disk)]
    [Guid("49275557-6329-4021-9215-3e7954ff8c25")]
    public class Windows7Libraries: ModuleTool
    {
        /* Всё устроено таким образом, что файл библиотеки загружается только на время 
         * проведения каких-либо манипуляций с ним. Т. е. даже когда на панели отображены
         * каталоги из какой-либо библиотеки, сама она свободна для использования извне.         
         *
         * Но для нормальной релизации такого дизайна пришлось завести костыль, 
         * который позволяет загрузать библиотеку только когда это требуется для 
         * обновления панелей, а не при каждом чихе пользователя. 
         * Для этого, все действия выполняемые над библиотекой (удаление\добавление каталогов,
         * и установка каталого по умолчанию) ставятся в очередь, и выполняются
         * когда в функции DisplayLibraryFolders(String) библиотека загружается для обновления
         * содержимого панели.
         * То есть библиотека загружается только тогда, когда по запросу фара нужно обновить 
         * содержимое панелей, и в этот же момент в библиотеку вносятся все запланированные изменения.
         * Запрос на перерисовку (вызов OnGettingData) обычно происходит каждый раз, сразу после того, 
         * как пользователь закрывает какой-нибудь подтветждающий дилог. */

        public const string Name = @"Windows7.Libraries";

        private IPanel _panel;
        private enum PanelState { libs, folders }; // состояние панели. устанавливается в OnSettingDirectory и используется в OnGettingData.
        private enum QueueMode { add, remove, setdef, skip };
        private struct MyPanel
        {
            public IPanel Panel;
            public String CurrentLibName; // имя текущей библиотеки
            public PanelState PanelState; // состяние панелей (библиотеки\каталоги)
            public QueueMode QueueMode; // состояние очереди обработки
            public List<FarFile> DirsQueue; // очередь обработки
        };
        MyPanel myPanel;

        public override void Invoke(object sender, ModuleToolEventArgs e)
        {
            if (!(Environment.OSVersion.Version.Major > 6 || (Environment.OSVersion.Version.Major == 6 && Environment.OSVersion.Version.Minor >= 1))) 
            {
                Far.Net.Message(GetString("MsgErrUnsuppOS"), Name, MsgOptions.Warning);
                return;
            }
            
            myPanel = new MyPanel();

            myPanel.Panel = Far.Net.CreatePanel();
            myPanel.Panel.AddDots = true;
            myPanel.Panel.Info.UseHighlighting = true;
            myPanel.Panel.Info.Title = Name;
            myPanel.Panel.Info.CurrentDirectory = Far.Net.Panel.Path;
            
            myPanel.Panel.DeletingFiles += OnDeletingLibrary; // удаление библиотеки или каталога из библиотеки
            myPanel.Panel.PuttingFiles += OnPuttingDirectory; // добавление каталогов в библиотеку
            myPanel.Panel.SettingDirectory += OnSettingDirectory; // смена вида: библиотеки, каталоги библиотеки, переход в каталог, выход
            myPanel.Panel.GettingData += OnGettingData; // перечитывание списка при обновлении панели
            myPanel.Panel.KeyPressed += OnKeyPressed; // создание библиотеки, переименование, назначение папки по умолчанию
            //myPanel.Panel.MakingDirectory += OnMakingLibrary;
            //myPanel.Panel.Closed += OnPanelClosed;
            
            myPanel.DirsQueue = new List<FarFile>();
            myPanel.QueueMode = QueueMode.skip;
            myPanel.CurrentLibName = null;
            
            this.KeyBarFolders = new string[] { "", "", "", GetString("KeySetDef"), "", "", "", "", "", "", "", "" };
            this.KeyBarLibs = new string[] { "", "", "", "", "", "", GetString("KeyNewLib"), "", "", "", "", "" };          
           
            // "особый" режим панели для отображения каталогов библиотеки и индикации установленного по умолчанию
            myPanel.Panel.Info.SetMode(
                PanelViewMode.Descriptions,
                new PanelModeInfo{
                    Columns = new FarColumn[]
                    {
                        new SetColumn{Kind = "N", Name = GetString("ColName"), Width = -30},
                        new SetColumn{Kind = "C0",Name = "*",                  Width = 1},
                        new SetColumn{Kind = "Z", Name = GetString("ColPath"), Width = 0}
                    }
                }
            );
            /* TODO: для отображения библиотек тоже можно сделать свой режим. 
             * отображать в нём колонку с количеством папок и путь к дефолной */

            _panel = myPanel.Panel; // берём псевдоним для сокращения

            _panel.Open();
            DisplayLibraries();
        }

        // срабатывает автоматически при обновлении панели
        private void OnGettingData(object sender, PanelEventArgs e)
        {
            switch (myPanel.PanelState) 
            { 
                case PanelState.libs:
                    if(myPanel.CurrentLibName != null)
                        DisplayLibraries(myPanel.CurrentLibName);
                    else if (Far.Net.Panel.CurrentFile != null)
                        DisplayLibraries(Far.Net.Panel.CurrentFile.Name);
                    else 
                        DisplayLibraries();
                    break;
                case PanelState.folders:
                    this.DisplayLibraryFolders(myPanel.CurrentLibName);
                    break;
            }            
        }

        // обрабатываем смену директории и выставляем режимы отображения (библиотеки / каталоги)
        private void OnSettingDirectory(object sender, SettingDirectoryEventArgs e)
        {   
            /* не знаю как как правильно и нужно ли обрабатывать эти режимы. поэтому отключил. */
            if (e.Mode == OperationModes.View ||
                e.Mode == OperationModes.QuickView ||
                e.Mode == OperationModes.Find ||
                e.Mode == OperationModes.Silent)
            {
                //Far.Message(e.Name, e.Mode.ToString());
                e.Ignore = true;
                return;
            }

            switch (e.Name)
            {   // переход на уровень вверх
                case "..":
                    // из списка каталогов в библиотеки
                    if (myPanel.PanelState == PanelState.folders)
                    {
                        // перелкючаем режим
                        myPanel.PanelState = PanelState.libs;
                        // а дальше, в OnGettingData, вызывается DisplayLibraryFolders();
                        break;
                    }
                    
                    // из библиотек в реальные панели
                    _panel.Close();
                    break;

                // выход в корень (CtrlBackSlash)
                case "\\":
                    _panel.Close();
                    break;

                // сюда бы заходил (быстрый)просмотр, если бы не был отключен
                default:
                    // вход в библиотеку
                    if (myPanel.PanelState == PanelState.libs)
                    {
                        myPanel.CurrentLibName = e.Name;
                        // переключаем режим
                        myPanel.PanelState = PanelState.folders;
                        // содержимое панели обновится после вызова OnGettingData
                        break;
                    }
                    //else if (myPanel.PanelState == PanelState.folders)
                    // вход в каталог библиотеки приводит закрытию панели плагина и переходу в этот каталог
                    _panel.Close(_panel.CurrentFile.AlternateName);                    
                    break;
            }
        }


        // срабатывает при нажатии F8 в панели
        private void OnDeletingLibrary(object sender, FilesEventArgs e)
        {
            string[] msgButtons1 = new string[2] { GetString("BtnDelete"), GetString("BtnCancel") };
            string[] msgButtons2 = new string[2] { GetString("BtnRemove"), GetString("BtnCancel") };

            if (myPanel.PanelState == PanelState.libs)
            {
                if (e.Files.Count == 1)
                {
                    if (0 != Far.Net.Message(String.Format(GetString("MsgConfDelLib"), e.Files[0].Name), Name, MsgOptions.None, msgButtons1))
                        return;
                }
                else
                {
                    if (0 != Far.Net.Message(String.Format(GetString("MsgConfDelSelLib"), e.Files.Count), Name, MsgOptions.None, msgButtons1))
                        return;
                }

                // удаляем
                foreach (FarFile file in e.Files)
                {
                    File.Delete(file.AlternateName);
                    _panel.Files.Remove(file);
                }
            }
            else
            {   // спрашиваем             
                if (e.Files.Count == 1)
                {
                    if (0 != Far.Net.Message(String.Format(GetString("MsgConfRemFldr"), e.Files[0].Name, e.Files[0].AlternateName), msgButtons2[0], MsgOptions.None, msgButtons2))
                        return;
                }
                else
                {
                    if (0 != Far.Net.Message(String.Format(GetString("MsgConfRemSelFldrs"), e.Files.Count), msgButtons2[0], MsgOptions.None, msgButtons2))
                        return;
                }

                // помещаем удаляемые каталоги в "очередь". реально они будут удалены при обновлении панели
                foreach (FarFile file in e.Files)
                {
                    myPanel.DirsQueue.Add(file);
                }
                myPanel.QueueMode = QueueMode.remove;
            }
            
        }

        // добавляем папку в библиотеку
        private void OnPuttingDirectory(object sender, PuttingFilesEventArgs e)
        {
            if (myPanel.PanelState == PanelState.libs)
            {
                Far.Net.Message(GetString("MakingLibraryHint"), Name, MsgOptions.Warning);
                return;
            }

            foreach (FarFile file in e.Files)
            {
                if (!file.IsDirectory)
                {
                    if (0 == Far.Net.Message(String.Format(GetString("MsgErrAddDir"), file.Name),
                          Name, MsgOptions.Warning,
                          new string[2] { GetString("BtnSkipIt"), GetString("BtnAbort") })
                        )
                        continue;
                    else
                        break;
                }

                // уже существующие просто игнориуем. типо добавили уже. лишний шум не поднимаем.
                if (ItemExistInPanel(file.Name))
                    continue;

                // добавляем каталоги в очередь
                file.AlternateName = Path.Combine(e.Source, file.Name);
                myPanel.DirsQueue.Add(file);
            }
            // не забываем установить режим обработки очереди
            myPanel.QueueMode = QueueMode.add;
        }

        // 
        private void OnKeyPressed(object sender, PanelKeyEventArgs e)
        {
            switch (e.Code)
            {   // делаем выделенную папку папкой по умолчанию
                case VKeyCode.F4:
                {                    
                    if (_panel.CurrentIndex == 0) return;
                    switch (e.State)
                    {
                        case KeyStates.None:
                            if (myPanel.PanelState == PanelState.folders)
                            {
                                myPanel.DirsQueue.Add(_panel.CurrentFile);
                                myPanel.QueueMode = QueueMode.setdef;
                                _panel.Update(true);
                                e.Ignore = true;
                            }
                            return;
                        default:
                            return;
                    }
                }
                // переименовываем библиотеку под курсором (выделение игнорируется)
                case VKeyCode.F6:
                {
                    switch (e.State)
                    {
                        case KeyStates.Shift:
                            if (myPanel.PanelState == PanelState.libs)
                                RenameLibrary();
                            e.Ignore = true; // а то чё-то два раза приходит
                            return;
                        default:
                            return;
                    }
                }
                // создаём новую библиотеку (и добавляем каталог?)
                case VKeyCode.F7:
                {
                    switch (e.State)
                    {
                        case KeyStates.None:
                            if (myPanel.PanelState == PanelState.libs)
                            {
                                CreateLibrary();
                            }
                            else
                            {   // TODO: выводить диалог и запрашивать путь к каталогу
                                Far.Net.Message(GetString("MakingFolderHint"), Name);
                            }
                            e.Ignore = true;
                            return;
                        default:
                            return;
                    }
                }
                
                //ToDo: хотекй на открытие библиотек в проводнике ::{031E4825-7B94-4DC3-B131-E946B44C8DD5}
                // e.ignore = true;
            }            
        }

        // создаём новую библиотеку
        private void CreateLibrary()
        {
            /*IsCorrectLibName checkNameFunc = delegate(string name) {
                return !ItemExistInPanel(name); 
            };*/
            string newLibName = String.Empty;
            if (GetCorrectLibName(ref newLibName, GetString("MsgMakeNewLibTitle"))/*, checkNameFunc*/)
            {
                new ShellLibrary(newLibName, false).Close();
                // делаем её не надолго текущей 
                myPanel.CurrentLibName = newLibName;
                _panel.Update(true); // приведёт к вызову DisplayLiraries, позицированию курсора, и занулению myPanel.CurrentLibName
            }
        }

        // переименовываем библиотеку (папки переименовывать нельзя)
        private void RenameLibrary()
        {
            if (_panel.CurrentIndex == 0) return;

            string oldName = GetLibraryPath(_panel.CurrentFile.Name); // исходное имя библиотеки
            string newLibName = _panel.CurrentFile.Name;

            /*IsCorrectLibName checkNameFunc = delegate(string name) {
                return !File.Exists(GetLibraryPath(name)); 
            };*/
            if (GetCorrectLibName(ref newLibName, String.Format(GetString("DlgRenLibTitle"), newLibName)/*, checkNameFunc*/))
            {
                string newName = GetLibraryPath(newLibName);
                File.Move(oldName, newName);
                _panel.CurrentFile.Name = newLibName;
                _panel.Update(false);
                _panel.GoToName(newLibName); // а если не двигать, то нужно вызывать _panel.Redraw() чтоб изменения отобразились
            }

        }

        // отображает библиотеки позицируя курсор на указанную
        private void DisplayLibraries(string nameToPost)
        {   //Far.Message(String.Format("DisplayLibraries({0})", nameToPost));            
            DisplayLibraries();
            _panel.PostName(nameToPost);
            // забываем использванную библиотеку
            myPanel.CurrentLibName = null;
        }

        // кладёт на пенель библиотеки (todo: сделать свой режим и отображать больше инфы)
        private void DisplayLibraries()
        {
            // получаем список библиотек из папки в которой их держит винда
            String[] libsFiles = Directory.GetFiles(ShellLibrary.LibrariesKnownFolder.Path, "*" + LibFileExtention);

            _panel.Files.Clear();
            _panel.ViewMode = PanelViewMode.Medium;
            for (int i = 0; i < libsFiles.Length; i++)
            {
                this.AddLibrary2Panel(Path.GetFileNameWithoutExtension(libsFiles[i]));
            }

            _panel.Info.SetKeyBarMain(KeyBarLibs);
            _panel.Info.Title = Name;
            // выставляем соответствующее состояние (где-то оно устанавливается ещё до вызова самой функции, но лишний раз внутри не мешает)
            myPanel.PanelState = PanelState.libs;
        }
        
        // добавлет на панель одну библиотеку
        private void AddLibrary2Panel(string LibraryName)
        {
            _panel.Files.Add(
                new SetFile
                {
                    Name = LibraryName,
                    AlternateName = GetLibraryPath(LibraryName),
                    IsDirectory = true
                }
            );
        }

        // загружает библиотеку, отрабатываем в ней изменения, и отображет её содерижмое на панели
        private void DisplayLibraryFolders(String libraryName)
        {
            ShellLibrary lib;
            LoadLibrary(libraryName, out lib, myPanel.QueueMode == QueueMode.skip); // "только для чтения" открываем только если очередь пуста

            // отложенная обработка изменений в библиотеке
            if (myPanel.QueueMode != QueueMode.skip)
            {
                foreach (FarFile folder in myPanel.DirsQueue)
                {
                    switch (myPanel.QueueMode) { 
                        case QueueMode.remove:
                            lib.Remove(folder.AlternateName); // BUG: невозможно удалить уже несуществующий путь
                            break;

                        case QueueMode.add:
                            lib.Add(folder.AlternateName);
                            break;

                        case QueueMode.setdef:                            
                            lib.DefaultSaveFolder = folder.AlternateName;
                            break;           
                    }
                }
                myPanel.DirsQueue.Clear();
                myPanel.QueueMode = QueueMode.skip;
            }            
            // отображение содержимого
            DisplayLibraryFolders(lib);
            lib.Close();
        }


        // кладёт на панель папки из библиотеки
        private void DisplayLibraryFolders(ShellLibrary library)
        {
            _panel.Files.Clear();
            _panel.ViewMode = PanelViewMode.Descriptions;

            for (int i = 0; i < library.Count; i++)
            {
                AddLibraryFolder2Panel(
                    library[i].Name,
                    library[i].Path, 
                    String.Equals(library[i].Path, library.DefaultSaveFolder, StringComparison.OrdinalIgnoreCase)
                );
            }

            myPanel.CurrentLibName = library.Name;
            myPanel.PanelState = PanelState.folders;
            _panel.Info.SetKeyBarMain(KeyBarFolders);
            _panel.Info.Title = String.Format("{0}\\{1}", Name, library.Name);
        }

        // добавляет одну папку на панель (с открытой библиотекой)
        private void AddLibraryFolder2Panel(string folderName, string altFolderName, bool isDefault)
        {
            SetFile folder = new SetFile
            {
                Name = folderName,
                AlternateName = altFolderName,
                Description = altFolderName,
                IsDirectory = true,
            };

            // дефолтную папку отмечаем звёздочкой в соответсвующей колонке
            if (isDefault)
                folder.Columns = new String[] { "*" };

            _panel.Files.Add(folder);
        }

        // пробует загрузить библиотеку. в случае чего ругается.
        bool LoadLibrary(string libName, out ShellLibrary lib, bool readOnly)
        {
            lib = ShellLibrary.Load(libName, readOnly);

            if (lib != null)
                return true;

            Far.Net.Message("Error on load library " + libName, Name, MsgOptions.Warning);
            return false;
        }


        // проверяет существование на панели файла с именем ItemName (используется только при добавлении папок в библиотеку)
        private bool ItemExistInPanel(string ItemName)
        {
            foreach (FarFile file in _panel.Files)
                if (file.Name.Equals(ItemName))
                    return true;
            return false;
        }

        // возвращает полный путь к файлу библиотеки с именем LibName
        public static string GetLibraryPath(string LibName)
        {
            return Path.Combine(ShellLibrary.LibrariesKnownFolder.Path, LibName + LibFileExtention);
        }

        // выпытывем у пользователя уникальное имя для библиотеки
        //public delegate bool IsCorrectLibName(string name); // всё упростилось и делегат стал не нужен
        public bool GetCorrectLibName(ref string name,  string promtTitle/*, IsCorrectLibName isCorrectName*/)
        {
            IInputBox ibox = Far.Net.CreateInputBox();
            ibox.EmptyEnabled = false;
            ibox.Prompt = GetString("DlgLibNamePromt");
            ibox.Title = promtTitle;
            ibox.Text = name;

            while (true)
            {
                if (ibox.Show())
                {
                    string newName = ibox.Text.Trim();
                    // имя не должно быть пустым и такая библиотека не должна уже существовать
                    if (!String.IsNullOrEmpty(newName) && !File.Exists(GetLibraryPath(newName))) //if (isCorrectName(newName))  
                    {
                        name = newName; // берём имя и заканчиваем
                        return true;
                    }
                    // если имя занято, то спрашиваем что делать (переименовать, отменить)
                    if (0 == Far.Net.Message(
                            String.Format(GetString("MsgLibExist"), newName),
                            promtTitle,
                            MsgOptions.None,
                            new string[2] {GetString("BtnRename"), GetString("BtnCancel")})
                        )
                    {
                        ibox.Text = newName;
                        continue; // справшиваем имя ещё раз
                    }

                    return false; // отменили диалог с вопросом о том, что делать
                }
                return false; // отменили диалог запроса имени
            }
        }

        //private void OnPanelClosed(object sender, EventArgs e) {  }

        
/*        
        private string Name
        {            
            get { return this.GetType().Name; }
            //get { return "Windows7.Libraries"; }
        }
*/
        private string[] KeyBarFolders;
        private string[] KeyBarLibs;
        
        private static string LibFileExtention = ".library-ms";    
    }
}
